/*
 * cluster.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*unit_t* find_barcode(unit_t *b)
{
    unsigned bi = (*b) >> INDEX_SHIFT;
    return bsearch(b,clustered_barcode_index[bi],volume_sizes[bi],cluster_entry_size,&compare_barcodes);
}*/

void free_volumes()
{
    if (index)
    {
        unsigned i;
        for (i=0; i<N_VOLUMES; i++) { if (index[i]) { free1(index[i],volume_allocated_sizes[i]); } }
        free1(index,sizeof(seq_t**)*N_VOLUMES); index = 0;
    }
    if (volume_allocated_sizes) { free1(volume_allocated_sizes,sizeof(*volume_allocated_sizes)*N_VOLUMES); volume_allocated_sizes = 0; }
}

void create_clusters()
{
    unsigned long long n = 0ull;
    cluster_entry_t *p;
    unsigned i,j;

    fprintf(stderr,"Creating clusters ..");

    clustered_barcode_index = (cluster_entry_t**)calloc1(N_VOLUMES,sizeof(cluster_entry_t*));
    cluster_entries = (cluster_entry_t*)malloc1(sizeof(cluster_entry_t)*n_unique_barcodes);
    p = cluster_entries;

    for (i=0; i<N_VOLUMES; i++)
    {
        clustered_barcode_index[i] = p;
        for (j=0; j<volume_sizes[i]; j++,n++,p++)
        {
            p->seq = index[i][j];
            p->next = p;
            p->cluster_id = n;
        }
    }
    assert(p == cluster_entries + n_unique_barcodes);

    free_volumes();
    fprintf(stderr," OK (now %lld bytes allocated)\n",allocated);
    /*fprintf(stderr," OK\n");*/
}

/*
 * Generates the next combination.
 * Returns 1 on success, 0 when there is no next combination.
 */
unsigned next_combination(unsigned *c,const unsigned n,const int k)
{
    int i;
    if (c[k-1] < n-1) { c[k-1]++; return 1; }
    for (i=k-2; i>=0; i--) { if (c[i] < n-k+i) { break; } }
    if (i < 0) { return 0; }
    c[i]++;
    for (; i<k-1; i++) { c[i+1] = c[i]+1; }
    return 1;
}

/*
 * Tries to merge two clusters, containing barcodes b1 and b2.
 * Returns: 1 if merged, 0 if s1 and s2 are already in the same cluster.
 */
unsigned merge_clusters(cluster_entry_t *e1,cluster_entry_t *e2)
{
    cluster_entry_t *from, *to;

    assert(e1);
    assert(e2);

    if (e1->cluster_id == e2->cluster_id) { return 0; }
    if (e1->cluster_id < e2->cluster_id) { from = e2; to = e1; } else { from = e1; to = e2; }

    if (from->next == from && to->next == to)
    {
        from->cluster_id = to->cluster_id;
        from->next = to;
        to->next = from;
    }
    else if (from->next == from)
    {
        from->cluster_id = to->cluster_id;
        from->next = to->next;
        to->next = from;
    }
    else if (to->next == to)
    {
        cluster_entry_t *t = from;
        do { t->cluster_id = to->cluster_id; t = t->next; } while (t != from);
        to->next = from->next;
        from->next = to;
    }
    else
    {
        cluster_entry_t *t = from;
        do { t->cluster_id = to->cluster_id; t = t->next; } while (t != from);
        t = from->next;
        from->next = to->next;
        to->next = t;
    }
    from->cluster_id = to->cluster_id;
    return 1;
}

#if 0
void cluster_by_mutating(unsigned d)
{
    int i;
    unit_t m[units_per_barcode];
    unsigned n_nuc_changes = 1 << (d*2);
    unsigned long long n_joined = 0ull;

    fprintf(stderr,"Clustering (with distance %u) ",d);
    for (i=0; i<n_unique_barcodes; i++)
    {
        unsigned p[d];
        unsigned nuc_change;
        unit_t *barcode = cluster_entries + cluster_entry_size*i;

        if (i%5000==0) { fprintf(stderr,"."); }
        /*printf("\n\n"); print_complete_barcode(cluster_entries + cluster_entry_size*i); printf("\n");*/

        { unsigned pi; for (pi=0; pi<d; pi++) { p[pi] = pi; } }
        do {
            for (nuc_change=0; nuc_change<n_nuc_changes; nuc_change++)
            {
                unsigned j;
                unsigned change = nuc_change;
                unit_t* f;

                copy_barcode(m,cluster_entries + cluster_entry_size*i);
                for (j=0; j<d; j++) { write_barcode_at_pos(m,p[j],change&3); change >>= 2; }
                if (compare_barcodes(barcode,m) >= 0) { continue; }
                /*print_complete_barcode(m); printf("\n");*/

                f = find_barcode(m);
                if (!f) { continue; }
                n_joined += merge_clusters(barcode,f);
            }
        } while (next_combination(p,barcode_complexity,d));
    }
    fprintf(stderr," OK (%llu joined)\n",n_joined);
}

void cluster_all_by_mutating()
{
    unsigned d;
    for (d=1; d<=dist; d++) { cluster_by_mutating(d); }
}
#endif

unsigned distance_between_indexes(unit_t a,unit_t b)
{
    unsigned d = 0u, i;
    for (i=INDEX_DEPTH; i; i--,a>>=2,b>>=2) { if ( (a & 3) != (b & 3) ) { d++; } }
    return d;
}

/*unsigned distance_between_barcodes(unit_t *pa,unit_t *pb)
{
    unsigned d = 0u, i;
    int k;
    assert(pa);
    assert(pb);
    for (k=0; k<units_per_barcode; k++,pa++,pb++)
    {
        unit_t a = *pa, b = *pb;
        for (i=BP_PER_UNIT; i; i--,a>>=2,b>>=2) { if ( (a & 3) != (b & 3) ) { d++; } }
    }
    return d;
}*/

unsigned distance_between_barcodes_is_d(unit_t *pa,unit_t *pb,unsigned dd)
{
    unsigned d = 0u, i, k;
    assert(pa);
    assert(pb);
    for (k=0; k<units_per_barcode; k++,pa++,pb++)
    {
        unit_t a = *pa, b = *pb;
        for (i=BP_PER_UNIT; i; i--,a>>=2,b>>=2) { if ( (a & 3) != (b & 3) ) { d++; if (d > dd) { return 0; } } }
    }
    return (d == dd) ? 1 : 0;
}

unsigned distance_between_seq_barcodes_is_d(seq_t *s1,seq_t *s2,unsigned d)
{
    return distance_between_barcodes_is_d((unit_t*)(s1+1),(unit_t*)(s2+1),d);
}

unsigned long long cluster_barcodes_of_two_volumes_by_comparing(unsigned d,unsigned long long index_1,unsigned long long index_2)
{
    unsigned long long n_joined = 0ull;
    cluster_entry_t *end1, *end2, *s1, *s2;

    end1 = (index_1 == (1ull << (INDEX_DEPTH * 2)) - 1) ? cluster_entries + n_unique_barcodes : clustered_barcode_index[index_1 + 1];
    end2 = (index_2 == (1ull << (INDEX_DEPTH * 2)) - 1) ? cluster_entries + n_unique_barcodes : clustered_barcode_index[index_2 + 1];

    for (s1=clustered_barcode_index[index_1]; s1<end1; s1++)
    {
        for (s2=clustered_barcode_index[index_2]; s2<end2; s2++)
        {
            if (distance_between_seq_barcodes_is_d(s1->seq,s2->seq,d)) { n_joined += merge_clusters(s1,s2); }
        }
    }

    return n_joined;
}

void cluster_by_comparing(unsigned d)
{
    unsigned long long n_joined = 0ull;
    unsigned index_dist;  /* How many differences in the index part of barcode. */

    fprintf(stderr,"Clustering with distance %u ..",d);
    /*printf("Clustering (with distance %u)\n",d);*/

    for (index_dist=0; index_dist<=d; index_dist++)
    {
        unsigned tail_dist;  /* How many differences in the rest of barcode. */
        if (index_dist > INDEX_DEPTH) { continue; }
        tail_dist = d - index_dist;
        if (tail_dist > barcode_complexity - INDEX_DEPTH) { continue; }

        /*printf("    === index distance: %u, tail distance: %u ===\n",index_dist,tail_dist);*/
        if (index_dist)
        {
            unsigned long long index_1;
            for (index_1 = 0ull; index_1 < (1ull << (INDEX_DEPTH * 2)); index_1++)
            {
                unsigned index_mutate_pos[index_dist];
                unsigned n_index_nuc_changes = 1 << (index_dist*2);
                /*printf("        comparing volume %llu\n",index_1);*/
                { unsigned pi; for (pi=0; pi<index_dist; pi++) { index_mutate_pos[pi] = pi; } }
                do {
                    unsigned index_nuc_change;
                    for (index_nuc_change = 0; index_nuc_change < n_index_nuc_changes; index_nuc_change++)
                    {
                        unsigned long long index_2 = index_1;
                        unsigned change = index_nuc_change;
                        unsigned j;
                        unsigned ok = 1;
                        for (j=0; j<index_dist; j++)
                        {
                            if (read_index_at_pos(index_2,index_mutate_pos[j]) == (change&3)) { ok = 0; break; }
                            write_index_at_pos(index_2,index_mutate_pos[j],change&3);
                            change >>= 2;
                        }
                        if (!ok) { continue; }
                        if (index_2 <= index_1) { continue; }
                        assert(distance_between_indexes(index_1,index_2) == index_dist);

                        /*printf("            with %llu",index_2);
                        printf(" (");
                        { unsigned ii; for (ii=0; ii<index_dist; ii++) { printf("%s%d",(ii>0?",":""),index_mutate_pos[ii]); } }
                        printf("), %u\n",index_nuc_change);*/

                        n_joined += cluster_barcodes_of_two_volumes_by_comparing(d,index_1,index_2);
                    }
                } while (next_combination(index_mutate_pos,INDEX_DEPTH,index_dist));

                /*unsigned long long index_2;
                for (index_2 = index_1; index_2 < (1ull << (INDEX_DEPTH * 2)); index_2++)
                {
                    if (distance_between_indexes(index_1,index_2) != d) { continue; }
                    printf("        comparing volumes %llu and %llu\n",index_1,index_2);
                    n_joined += cluster_barcodes_of_two_volumes_by_comparing(d,index_1,index_2);
                }*/
            }
        }
        else
        {
            unsigned long long index_1;
            for (index_1 = 0ull; index_1 < (1ull << (INDEX_DEPTH * 2)); index_1++)
            {
                /*printf("        comparing volumes %llu and %llu\n",index_1,index_1);*/
                n_joined += cluster_barcodes_of_two_volumes_by_comparing(d,index_1,index_1);
            }
        }
    }
    fprintf(stderr," OK, %llu joined\n",n_joined);
    /*printf("Clustering complete (%llu joined)\n",n_joined);*/
}

void cluster_all_by_comparing()
{
    unsigned d;
    for (d=1; d<=dist; d++) { cluster_by_comparing(d); }
}

void cluster_barcodes()
{
    create_clusters();
    /*cluster_all_by_mutating();*/
    cluster_all_by_comparing();
}

#endif
