/*
 * globals.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

static unsigned is_nucleotide[256];   /* Used to detect nucleotide codes, and to convert them to binary form (a -> 1, c -> 2, g -> 3, t -> 4). */
static const char binary_to_nucleotide[4] = {'A','C','G','T'};  /* Converts binary representation back to nucleotide character. */


static long long allocated = 0ll;      /* How many bytes are allocated currently. */
static long long max_allocated = 0ll;  /* Maximum amount of memory that was allocated in current program run. */

static int opt_help = 0;     /* Flag set by --help option. */
static int opt_version = 0;  /* Flag set by --version option. */

static char *dist_str = 0;  /* Distance used for clustering. */
static unsigned dist = 0;

static char *template_parameter = 0;  /* Barcode template, specified with --template option. */
static char *template = 0;
static unsigned template_length = 0;

static char *use_barcode_position = 0;   /* Specifies which barcode positions should be used (based on template). */
static unsigned barcode_complexity = 0;  /* Number of meaningful positions in barcode. It's the number of N's in template. */

static unsigned units_per_barcode = 0;  /* Number of units and bytes necessary to store one barcode in binary represenatation. */
static unsigned bytes_per_barcode = 0;  /* With 64-bit units, one unit stores 32 bp of barcode. */



static char *input_path = 0;   /* Input file, specified with --in option. */
static char *output_path = 0;  /* Output file, specified with --out option. */
static FILE* OUT = 0;

static char *input = 0;                    /* The whole input gets loaded here. */
static unsigned long long input_size = 0;  /* Size of the input, in bytes. */

/* A sequence - points to the data within input. */
/* The actual structure also contains binary barcode at the end. */
typedef struct seq_t { char *name, *data; } seq_t;
static unsigned long long seq_size = 0;  /* Size of one sequence (seq_t), including barcode. */

static seq_t *seqs = 0;                   /* Sequences. */
static unsigned long long n_seqs = 0;     /* Number of sequences. */
static unsigned long long n_seqs_ok = 0;  /* Number of sequences with barcodes that match template. */



/* Sequence addresses stored in volumes for quick sorting and removing duplicates. */
static seq_t ***index = 0;
static unsigned long long *volume_allocated_sizes = 0;
static unsigned long long *volume_sizes = 0;


static unsigned long long n_unique_barcodes = 0ll;

typedef struct cluster_entry_t { seq_t *seq; void *next; long long cluster_id; } cluster_entry_t;
static cluster_entry_t *cluster_entries = 0;
static cluster_entry_t **clustered_barcode_index = 0;

#endif
